// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::AluBranchMOp,
    unit::{DynUnit, DynUnitWrapper, UnitCancelInput, UnitKind, UnitOutput, UnitTrait},
};
use fayalite::{
    intern::{Intern, Interned},
    prelude::*,
    util::ready_valid::ReadyValid,
};

#[hdl_module]
pub fn alu_branch(config: &CpuConfig) {
    #[hdl]
    let cd: ClockDomain = m.input();
    // TODO: finish
}

#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
pub struct AluBranch {
    config: Interned<CpuConfig>,
    module: Interned<Module<alu_branch>>,
}

impl AluBranch {
    pub fn new(config: &CpuConfig) -> Self {
        Self {
            config: config.intern(),
            module: alu_branch(config),
        }
    }
}

impl UnitTrait for AluBranch {
    type Type = alu_branch;
    type ExtraOut = ();
    type MOp = AluBranchMOp<DynSize>;

    fn ty(&self) -> Self::Type {
        self.module.io_ty()
    }

    fn extra_out_ty(&self) -> Self::ExtraOut {
        ()
    }

    fn mop_ty(&self) -> Self::MOp {
        AluBranchMOp[self.config.p_reg_num().canonical().bit_width()]
    }

    fn unit_kind(&self) -> UnitKind {
        UnitKind::AluBranch
    }

    fn make_module(&self) -> Interned<Module<Self::Type>> {
        self.module
    }

    fn cancel_input(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitCancelInput<DynSize, DynSize>>> {
        todo!()
    }

    fn output(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitOutput<DynSize, DynSize, Self::ExtraOut>>> {
        todo!()
    }

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain> {
        this.cd
    }

    fn to_dyn(&self) -> DynUnit {
        DynUnitWrapper(*self).to_dyn()
    }
}
